﻿using System;
using System.Collections.Generic;
using System.Net;
using System.Threading;
using System.Web;
using System.Web.Services.Configuration;
using System.Web.Services.Description;
using System.Web.Services.Protocols;
using System.Xml;

namespace MySoapLib {
    internal class Soap11ServerProtocolHelper : SoapServerProtocolHelper {
        // Methods
        internal Soap11ServerProtocolHelper(SoapServerProtocol protocol)
            : base(protocol) {
        }

        internal Soap11ServerProtocolHelper(SoapServerProtocol protocol, string requestNamespace)
            : base(protocol, requestNamespace) {
        }

        internal override SoapServerMethod RouteRequest() {
            object requestElement;
            string str = base.ServerProtocol.Request.Headers["SOAPAction"];
            if (str == null) {
                throw new SoapException(Res.GetString("UnableToHandleRequestActionRequired0"), new XmlQualifiedName("Client", "http://schemas.xmlsoap.org/soap/envelope/"));
            }
            if (base.ServerType.routingOnSoapAction) {
                if (str.StartsWith("\"", StringComparison.Ordinal) && str.EndsWith("\"", StringComparison.Ordinal)) {
                    str = str.Substring(1, str.Length - 2);
                }
                requestElement = HttpUtility.UrlDecode(str);
            }
            else {
                try {
                    requestElement = base.GetRequestElement();
                }
                catch (SoapException) {
                    throw;
                }
                catch (Exception exception) {
                    if (((exception is ThreadAbortException) || (exception is StackOverflowException)) || (exception is OutOfMemoryException)) {
                        throw;
                    }
                    throw new SoapException(Res.GetString("TheRootElementForTheRequestCouldNotBeDetermined0"), new XmlQualifiedName("Server", "http://schemas.xmlsoap.org/soap/envelope/"), exception);
                }
                catch {
                    throw new SoapException(Res.GetString("TheRootElementForTheRequestCouldNotBeDetermined0"), new XmlQualifiedName("Server", "http://schemas.xmlsoap.org/soap/envelope/"), new Exception());
                }
            }
           
            SoapServerMethod method = base.ServerType.GetMethod(requestElement);
        
            if (method != null) {
                return method;
            }
            if (base.ServerType.routingOnSoapAction) {
                throw new SoapException(Res.GetString("WebHttpHeader", new object[] { "SOAPAction", (string)requestElement }), new XmlQualifiedName("Client", "http://schemas.xmlsoap.org/soap/envelope/"));
            }
            throw new SoapException(Res.GetString("TheRequestElementXmlnsWasNotRecognized2", new object[] { ((XmlQualifiedName)requestElement).Name, ((XmlQualifiedName)requestElement).Namespace }), new XmlQualifiedName("Client", "http://schemas.xmlsoap.org/soap/envelope/"));
        }

        private static XmlQualifiedName TranslateFaultCode(XmlQualifiedName code) {
            if (code.Namespace != "http://schemas.xmlsoap.org/soap/envelope/") {
                if (!(code.Namespace == "http://www.w3.org/2003/05/soap-envelope")) {
                    return code;
                }
                if (code.Name == "Receiver") {
                    return SoapException.ServerFaultCode;
                }
                if (code.Name == "Sender") {
                    return SoapException.ClientFaultCode;
                }
                if (code.Name == "MustUnderstand") {
                    return SoapException.MustUnderstandFaultCode;
                }
                if (code.Name == "VersionMismatch") {
                    return SoapException.VersionMismatchFaultCode;
                }
            }
            return code;
        }

        internal override void WriteFault(XmlWriter writer, SoapException soapException, HttpStatusCode statusCode) {
            if ((statusCode == HttpStatusCode.InternalServerError) && (soapException != null)) {
                SoapServerMessage message = base.ServerProtocol.Message;
                writer.WriteStartDocument();
                writer.WriteStartElement("soap", "Envelope", "http://schemas.xmlsoap.org/soap/envelope/");
                writer.WriteAttributeString("xmlns", "soap", null, "http://schemas.xmlsoap.org/soap/envelope/");
                writer.WriteAttributeString("xmlns", "xsi", null, "http://www.w3.org/2001/XMLSchema-instance");
                writer.WriteAttributeString("xmlns", "xsd", null, "http://www.w3.org/2001/XMLSchema");
                if (base.ServerProtocol.ServerMethod != null) {
                    SoapHeaderHandling.WriteHeaders(writer, base.ServerProtocol.ServerMethod.outHeaderSerializer, message.Headers, base.ServerProtocol.ServerMethod.outHeaderMappings, SoapHeaderDirection.Fault, base.ServerProtocol.ServerMethod.use == SoapBindingUse.Encoded, base.ServerType.serviceNamespace, base.ServerType.serviceDefaultIsEncoded, "http://schemas.xmlsoap.org/soap/envelope/");
                }
                else {
                    SoapHeaderHandling.WriteUnknownHeaders(writer, message.Headers, "http://schemas.xmlsoap.org/soap/envelope/");
                }
                writer.WriteStartElement("Body", "http://schemas.xmlsoap.org/soap/envelope/");
                writer.WriteStartElement("Fault", "http://schemas.xmlsoap.org/soap/envelope/");
                writer.WriteStartElement("faultcode", "");
                XmlQualifiedName name = TranslateFaultCode(soapException.Code);
                if (((name.Namespace != null) && (name.Namespace.Length > 0)) && (writer.LookupPrefix(name.Namespace) == null)) {
                    writer.WriteAttributeString("xmlns", "q0", null, name.Namespace);
                }
                writer.WriteQualifiedName(name.Name, name.Namespace);
                writer.WriteEndElement();
                writer.WriteStartElement("faultstring", "");
                if ((soapException.Lang != null) && (soapException.Lang.Length != 0)) {
                    writer.WriteAttributeString("xml", "lang", "http://www.w3.org/XML/1998/namespace", soapException.Lang);
                }
                writer.WriteString(base.ServerProtocol.GenerateFaultString(soapException));
                writer.WriteEndElement();
                string actor = soapException.Actor;
                if (actor.Length > 0) {
                    writer.WriteElementString("faultactor", "", actor);
                }
                if (!(soapException is SoapHeaderException)) {
                    if (soapException.Detail == null) {
                        writer.WriteStartElement("detail", "");
                        writer.WriteEndElement();
                    }
                    else {
                        soapException.Detail.WriteTo(writer);
                    }
                }
                writer.WriteEndElement();
                writer.WriteEndElement();
                writer.WriteEndElement();
                writer.Flush();
            }
        }

        // Properties
        internal override string EncodingNs {
            get {
                return "http://schemas.xmlsoap.org/soap/encoding/";
            }
        }

        internal override string EnvelopeNs {
            get {
                return "http://schemas.xmlsoap.org/soap/envelope/";
            }
        }

        internal override string HttpContentType {
            get {
                return "text/xml";
            }
        }

        internal override WebServiceProtocols Protocol {
            get {
                return WebServiceProtocols.HttpSoap;
            }
        }

        internal override SoapProtocolVersion Version {
            get {
                return SoapProtocolVersion.Soap11;
            }
        }
    }


    internal class Soap12ServerProtocolHelper : SoapServerProtocolHelper {
        // Methods
        internal Soap12ServerProtocolHelper(SoapServerProtocol protocol)
            : base(protocol) {
        }

        internal Soap12ServerProtocolHelper(SoapServerProtocol protocol, string requestNamespace)
            : base(protocol, requestNamespace) {
        }

        internal override SoapServerMethod RouteRequest() {
            string action = ContentType.GetAction(base.ServerProtocol.Request.ContentType);
            SoapServerMethod method = null;
            bool flag = false;
            bool flag2 = false;
           
            if ((action != null) && (action.Length > 0)) {
                action = HttpUtility.UrlDecode(action);
               
                method = base.ServerType.GetMethod(action);
               
                if ((method != null) && (base.ServerType.GetDuplicateMethod(action) != null)) {
                    method = null;
                    flag = true;
                }
            }
            XmlQualifiedName empty = XmlQualifiedName.Empty;
            if (method == null) {
                empty = base.GetRequestElement();
               
                method = base.ServerType.GetMethod(empty);
                
                if ((method != null) && (base.ServerType.GetDuplicateMethod(empty) != null)) {
                    method = null;
                    flag2 = true;
                }
            }
            if (method != null) {
                return method;
            }
            if ((action == null) || (action.Length == 0)) {
                throw new SoapException(Res.GetString("UnableToHandleRequestActionRequired0"), Soap12FaultCodes.SenderFaultCode);
            }
            if (flag) {
                if (flag2) {
                    throw new SoapException(Res.GetString("UnableToHandleRequest0"), Soap12FaultCodes.ReceiverFaultCode);
                }
                throw new SoapException(Res.GetString("TheRequestElementXmlnsWasNotRecognized2", new object[] { empty.Name, empty.Namespace }), Soap12FaultCodes.SenderFaultCode);
            }
            throw new SoapException(Res.GetString("UnableToHandleRequestActionNotRecognized1", new object[] { action }), Soap12FaultCodes.SenderFaultCode);
        }

        private static XmlQualifiedName TranslateFaultCode(XmlQualifiedName code) {
            if (code.Namespace == "http://schemas.xmlsoap.org/soap/envelope/") {
                if (code.Name == "Server") {
                    return Soap12FaultCodes.ReceiverFaultCode;
                }
                if (code.Name == "Client") {
                    return Soap12FaultCodes.SenderFaultCode;
                }
                if (code.Name == "MustUnderstand") {
                    return Soap12FaultCodes.MustUnderstandFaultCode;
                }
                if (code.Name == "VersionMismatch") {
                    return Soap12FaultCodes.VersionMismatchFaultCode;
                }
            }
            return code;
        }

        internal override void WriteFault(XmlWriter writer, SoapException soapException, HttpStatusCode statusCode) {
            if ((statusCode == HttpStatusCode.InternalServerError) && (soapException != null)) {
                writer.WriteStartDocument();
                writer.WriteStartElement("soap", "Envelope", "http://www.w3.org/2003/05/soap-envelope");
                writer.WriteAttributeString("xmlns", "soap", null, "http://www.w3.org/2003/05/soap-envelope");
                writer.WriteAttributeString("xmlns", "xsi", null, "http://www.w3.org/2001/XMLSchema-instance");
                writer.WriteAttributeString("xmlns", "xsd", null, "http://www.w3.org/2001/XMLSchema");
                if (base.ServerProtocol.ServerMethod != null) {
                    SoapHeaderHandling.WriteHeaders(writer, base.ServerProtocol.ServerMethod.outHeaderSerializer, base.ServerProtocol.Message.Headers, base.ServerProtocol.ServerMethod.outHeaderMappings, SoapHeaderDirection.Fault, base.ServerProtocol.ServerMethod.use == SoapBindingUse.Encoded, base.ServerType.serviceNamespace, base.ServerType.serviceDefaultIsEncoded, "http://www.w3.org/2003/05/soap-envelope");
                }
                else {
                    SoapHeaderHandling.WriteUnknownHeaders(writer, base.ServerProtocol.Message.Headers, "http://www.w3.org/2003/05/soap-envelope");
                }
                writer.WriteStartElement("Body", "http://www.w3.org/2003/05/soap-envelope");
                writer.WriteStartElement("Fault", "http://www.w3.org/2003/05/soap-envelope");
                writer.WriteStartElement("Code", "http://www.w3.org/2003/05/soap-envelope");
                WriteFaultCodeValue(writer, TranslateFaultCode(soapException.Code), soapException.SubCode);
                writer.WriteEndElement();
                writer.WriteStartElement("Reason", "http://www.w3.org/2003/05/soap-envelope");
                writer.WriteStartElement("Text", "http://www.w3.org/2003/05/soap-envelope");
                writer.WriteAttributeString("xml", "lang", "http://www.w3.org/XML/1998/namespace", Res.GetString("XmlLang"));
                writer.WriteString(base.ServerProtocol.GenerateFaultString(soapException));
                writer.WriteEndElement();
                writer.WriteEndElement();
                string actor = soapException.Actor;
                if (actor.Length > 0) {
                    writer.WriteElementString("Node", "http://www.w3.org/2003/05/soap-envelope", actor);
                }
                string role = soapException.Role;
                if (role.Length > 0) {
                    writer.WriteElementString("Role", "http://www.w3.org/2003/05/soap-envelope", role);
                }
                if (!(soapException is SoapHeaderException)) {
                    if (soapException.Detail == null) {
                        writer.WriteStartElement("Detail", "http://www.w3.org/2003/05/soap-envelope");
                        writer.WriteEndElement();
                    }
                    else {
                        soapException.Detail.WriteTo(writer);
                    }
                }
                writer.WriteEndElement();
                writer.WriteEndElement();
                writer.WriteEndElement();
                writer.Flush();
            }
        }

        private static void WriteFaultCodeValue(XmlWriter writer, XmlQualifiedName code, SoapFaultSubCode subcode) {
            if (code != null) {
                writer.WriteStartElement("Value", "http://www.w3.org/2003/05/soap-envelope");
                if (((code.Namespace != null) && (code.Namespace.Length > 0)) && (writer.LookupPrefix(code.Namespace) == null)) {
                    writer.WriteAttributeString("xmlns", "q0", null, code.Namespace);
                }
                writer.WriteQualifiedName(code.Name, code.Namespace);
                writer.WriteEndElement();
                if (subcode != null) {
                    writer.WriteStartElement("Subcode", "http://www.w3.org/2003/05/soap-envelope");
                    WriteFaultCodeValue(writer, subcode.Code, subcode.SubCode);
                    writer.WriteEndElement();
                }
            }
        }

        // Properties
        internal override string EncodingNs {
            get {
                return "http://www.w3.org/2003/05/soap-encoding";
            }
        }

        internal override string EnvelopeNs {
            get {
                return "http://www.w3.org/2003/05/soap-envelope";
            }
        }

        internal override string HttpContentType {
            get {
                return "application/soap+xml";
            }
        }

        internal override WebServiceProtocols Protocol {
            get {
                return WebServiceProtocols.HttpSoap12;
            }
        }

        internal override SoapProtocolVersion Version {
            get {
                return SoapProtocolVersion.Soap12;
            }
        }
    }


    internal abstract class SoapServerProtocolHelper {
        // Fields
        private SoapServerProtocol protocol;
        private string requestNamespace;

        // Methods
        protected SoapServerProtocolHelper(SoapServerProtocol protocol) {
            this.protocol = protocol;
        }

        protected SoapServerProtocolHelper(SoapServerProtocol protocol, string requestNamespace) {
            this.protocol = protocol;
            this.requestNamespace = requestNamespace;
        }

        internal static SoapServerProtocolHelper GetHelper(SoapServerProtocol protocol, string envelopeNs) {
            if ((envelopeNs != "http://schemas.xmlsoap.org/soap/envelope/") && (envelopeNs == "http://www.w3.org/2003/05/soap-envelope")) {
                return new Soap12ServerProtocolHelper(protocol, envelopeNs);
            }
            return new Soap11ServerProtocolHelper(protocol, envelopeNs);
        }

        protected XmlQualifiedName GetRequestElement() {
            XmlQualifiedName empty;
            SoapServerMessage message = this.ServerProtocol.Message;
            long position = message.Stream.Position;
            XmlReader xmlReader = this.protocol.GetXmlReader();
            xmlReader.MoveToContent();
            this.requestNamespace = xmlReader.NamespaceURI;
            if (!xmlReader.IsStartElement("Envelope", this.requestNamespace)) {
                throw new InvalidOperationException(Res.GetString("WebMissingEnvelopeElement"));
            }
            if (xmlReader.IsEmptyElement) {
                throw new InvalidOperationException(Res.GetString("WebMissingBodyElement"));
            }
            xmlReader.ReadStartElement("Envelope", this.requestNamespace);
            xmlReader.MoveToContent();
            while (!xmlReader.EOF && !xmlReader.IsStartElement("Body", this.requestNamespace)) {
                xmlReader.Skip();
            }
            if (xmlReader.EOF) {
                throw new InvalidOperationException(Res.GetString("WebMissingBodyElement"));
            }
            if (xmlReader.IsEmptyElement) {
                empty = XmlQualifiedName.Empty;
            }
            else {
                xmlReader.ReadStartElement("Body", this.requestNamespace);
                xmlReader.MoveToContent();
                empty = new XmlQualifiedName(xmlReader.LocalName, xmlReader.NamespaceURI);
            }
            message.Stream.Position = position;
            return empty;
        }

        internal abstract SoapServerMethod RouteRequest();
        internal HttpStatusCode SetResponseErrorCode(HttpResponse response, SoapException soapException) {
            if ((soapException.SubCode != null) && (soapException.SubCode.Code == Soap12FaultCodes.UnsupportedMediaTypeFaultCode)) {
                response.StatusCode = 0x19f;
                soapException.ClearSubCode();
            }
            else if (SoapException.IsClientFaultCode(soapException.Code)) {
                SoapServerProtocol.SetHttpResponseStatusCode(response, 500);
                for (Exception exception = soapException; exception != null; exception = exception.InnerException) {
                    if (exception is XmlException) {
                        response.StatusCode = 400;
                    }
                }
            }
            else {
                SoapServerProtocol.SetHttpResponseStatusCode(response, 500);
            }
            response.StatusDescription = HttpWorkerRequest.GetStatusDescription(response.StatusCode);
            return (HttpStatusCode)response.StatusCode;
        }

        internal abstract void WriteFault(XmlWriter writer, SoapException soapException, HttpStatusCode statusCode);

        // Properties
        internal abstract string EncodingNs { get; }

        internal abstract string EnvelopeNs { get; }

        internal abstract string HttpContentType { get; }

        internal abstract WebServiceProtocols Protocol { get; }

        internal string RequestNamespace {
            get {
                return this.requestNamespace;
            }
        }

        protected SoapServerProtocol ServerProtocol {
            get {
                return this.protocol;
            }
        }

        protected SoapServerType ServerType {
            get {
                return (SoapServerType)this.protocol.ServerType;
            }
        }

        internal abstract SoapProtocolVersion Version { get; }
    }


}
